use core::fmt::Display;
use serenity::{
    builder::EditMember,
    http::{client::Http, CacheHttp},
    model::{
        channel::{Message, Reaction, ReactionType},
        guild::Member,
        id::{ChannelId, RoleId},
    },
};

pub trait LoggingReaction {
    fn try_delete(&self, cache_http: impl CacheHttp);
}

impl LoggingReaction for Reaction {
    fn try_delete(&self, cache_http: impl CacheHttp) {
        self.delete(cache_http)
            .unwrap_or_else(|e| error!("Unable to delete react: {:?}", e));
    }
}

pub trait LoggingMessage {
    fn try_react<R: Into<ReactionType>>(&self, cache_http: impl CacheHttp, reaction_type: R);
    fn try_delete(&self, cache_http: impl CacheHttp);
}

impl LoggingMessage for Message {
    fn try_react<R: Into<ReactionType>>(&self, cache_http: impl CacheHttp, reaction_type: R) {
        self.react(cache_http, reaction_type)
            .unwrap_or_else(|e| error!("Unable to add reaction, {:?}", e));
    }
    fn try_delete(&self, cache_http: impl CacheHttp) {
        self.delete(cache_http)
            .unwrap_or_else(|e| error!("Unable to delete message: {:?}", e));
    }
}

pub trait LoggingChannelId {
    fn try_say(self, http: impl AsRef<Http>, content: impl Display);
}

impl LoggingChannelId for ChannelId {
    fn try_say(self, http: impl AsRef<Http>, content: impl Display) {
        if let Err(e) = self.say(http, content) {
            error!("Error sending message {:?}", e)
        }
    }
}

pub trait LoggingMember {
    fn try_edit<F: FnOnce(&mut EditMember) -> &mut EditMember>(&self, http: impl AsRef<Http>, f: F);
    fn try_remove_role<R: Into<RoleId>>(&mut self, http: impl AsRef<Http>, role_id: R);
}

impl LoggingMember for Member {
    fn try_edit<F: FnOnce(&mut EditMember) -> &mut EditMember>(
        &self,
        http: impl AsRef<Http>,
        f: F,
    ) {
        self.edit(http, f)
            .unwrap_or_else(|e| error!("Unable to edit member: {:?}", e));
    }
    fn try_remove_role<R: Into<RoleId>>(&mut self, http: impl AsRef<Http>, role_id: R) {
        self.remove_role(http, role_id)
            .unwrap_or_else(|e| error!("Unable to remove role: {:?}", e));
    }
}

